/*$Id: ex7.c,v 1.31 2000/09/22 20:46:14 bsmith Exp $*/

static char help[] = "Solves a nonlinear system in parallel with SNES.\n\
We solve the driven cavity problem in a streamfunction-vorticity formulation.\n\
  -mx <xg>, where <xg> = number of grid points in the x-direction\n\
  -my <yg>, where <yg> = number of grid points in the y-direction\n\
  -Nx <npx>, where <npx> = number of processors in the x-direction\n\
  -Ny <npy>, where <npy> = number of processors in the y-direction\n\n";

/*T
   Concepts: SNES^parallel multicomponent example;
   Concepts: multicomponent;
   Concepts: DA^using distributed arrays;
   Concepts: structured grid
   Processors: n
T*/

/* ------------------------------------------------------------------------

    THIS CODE DOES NOT CURRENTLY WORK!

    This problem is modeled by the partial differential equation
  
  
    A finite difference approximation with the usual stencil is used to
    discretize the boundary value problem to obtain a nonlinear system of
    equations.

    The Jacobian can be 
      * formed via finite differencing using coloring (the default)
      * applied matrix-free via the option -snes_mf 
          (for larger grid problems this variant  may not converge due 
           to ill-conditioning of the matrix)
      * formed via an analytic formula (not yet fully implemented)

  ------------------------------------------------------------------------- */

/* 
   Include "petscda.h" so that we can use distributed arrays (DAs).
   Include "petscsnes.h" so that we can use SNES solvers.  Note that this
   file automatically includes:
     petsc.h       - base PETSc routines   petscvec.h - vectors
     petscsys.h    - system routines       petscmat.h - matrices
     petscis.h     - index sets            petscksp.h - Krylov subspace methods
     petscviewer.h - viewers               petscpc.h  - preconditioners
     petscsles.h   - linear solvers
*/
#include "petscsnes.h"
#include "petscda.h"

/* 
   User-defined application context - contains data needed by the 
   application-provided call-back routines, FormJacobian() and
   FormFunction().
*/
typedef struct {
   double      lidvelocity;         
   int         mx,my;          /* discretization in x, y directions */
   Vec         localX,localF; /* ghosted local vector */
   DA          da;             /* distributed array data structure */
   int         rank;           /* processor rank */
} AppCtx;

/* 
   User-defined routines
*/
extern int FormInitialGuess(AppCtx*,Vec);
extern int FormFunctionFast(SNES,Vec,Vec,void*);
extern int FormFunctionSlow(SNES,Vec,Vec,void*);
extern int FormJacobian(SNES,Vec,Mat*,Mat*,MatStructure*,void*);
extern int DAGetColoringSpecial(DA,ISColoring *,Mat *);

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **argv)
{
  SNES          snes;                /* nonlinear solver */
  Vec           x,r;                /* solution, residual vectors */
  Mat           J;                   /* Jacobian matrix */
  AppCtx        user;                /* user-defined work context */
  int           its;                 /* iterations for convergence */
  int           Nx,Ny;              /* number of preocessors in x- and y- directions */
  int           size,ierr;           /* number of processors */
  MatFDColoring fdcoloring;          /* matrix coloring context */
  PetscTruth    flg;


  /* Nonlinear function evaluation routine.  This example supports two
     variants: FormFunctionFast() and FormFunctionSlow(). */
  int           (*FormFunction)(SNES,Vec,Vec,void*) = FormFunctionFast;

  PetscInitialize(&argc,&argv,(char *)0,help);
  ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&user.rank);CHKERRA(ierr);

  /*
     Initialize problem parameters
  */
  user.mx = 4; user.my = 4; 
  ierr = OptionsGetInt(PETSC_NULL,"-mx",&user.mx,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-my",&user.my,PETSC_NULL);CHKERRA(ierr);

  user.lidvelocity = 1.0/(user.mx*user.my);
  ierr = OptionsGetDouble(PETSC_NULL,"-lidvelocity",&user.lidvelocity,PETSC_NULL);CHKERRA(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create nonlinear solver context
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = SNESCreate(PETSC_COMM_WORLD,SNES_NONLINEAR_EQUATIONS,&snes);CHKERRA(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create vector data structures; set function evaluation routine
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
     Create distributed array (DA) to manage parallel grid and vectors
  */
  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&size);CHKERRA(ierr);
  Nx = PETSC_DECIDE; Ny = PETSC_DECIDE;
  ierr = OptionsGetInt(PETSC_NULL,"-Nx",&Nx,PETSC_NULL);CHKERRA(ierr);
  ierr = OptionsGetInt(PETSC_NULL,"-Ny",&Ny,PETSC_NULL);CHKERRA(ierr);
  ierr = DACreate2d(PETSC_COMM_WORLD,DA_NONPERIODIC,DA_STENCIL_STAR,user.mx,
                    user.my,Nx,Ny,2,1,PETSC_NULL,PETSC_NULL,&user.da);CHKERRA(ierr);

  /*
     Extract global and local vectors from DA; then duplicate for remaining
     vectors that are the same types
  */
  ierr = DACreateGlobalVector(user.da,&x);CHKERRA(ierr);
  ierr = DACreateLocalVector(user.da,&user.localX);CHKERRA(ierr);
  ierr = VecDuplicate(x,&r);CHKERRA(ierr);
  ierr = VecDuplicate(user.localX,&user.localF);CHKERRA(ierr);

  /* 
     Set function evaluation routine and vector
  */
  ierr = OptionsHasName(PETSC_NULL,"-slow_function",&flg);
  if (flg) {
    FormFunction = FormFunctionSlow;
  }
  ierr = SNESSetFunction(snes,r,FormFunction,&user);CHKERRA(ierr);


  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create matrix data structure; set Jacobian evaluation routine
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
       Set up coloring information needed for sparse finite difference
       approximation of the Jacobian
   */
  {
  ISColoring iscoloring;
  ierr = DAGetColoringSpecial(user.da,&iscoloring,&J);CHKERRA(ierr);
  ierr = MatFDColoringCreate(J,iscoloring,&fdcoloring);CHKERRA(ierr); 
  ierr = MatFDColoringSetFunction(fdcoloring,(int (*)(void))FormFunction,&user);CHKERRA(ierr);
  ierr = MatFDColoringSetFromOptions(fdcoloring);CHKERRA(ierr); 
  ierr = ISColoringDestroy(iscoloring);CHKERRA(ierr);
  }

  /* 
     Set Jacobian matrix data structure and default Jacobian evaluation
     routine. User can override with:

     -snes_mf : matrix-free Newton-Krylov method with no preconditioning
                (unless user explicitly sets preconditioner) 
     -snes_mf_operator : form preconditioning matrix as set by the user,
                         but use matrix-free approx for Jacobian-vector
                         products within Newton-Krylov method

  */
  ierr = OptionsHasName(PETSC_NULL,"-analytic_jacobian",&flg);CHKERRA(ierr);
  if (flg) {
    ierr = SNESSetJacobian(snes,J,J,FormJacobian,&user);CHKERRA(ierr);
  } else {
    ierr = SNESSetJacobian(snes,J,J,SNESDefaultComputeJacobianColor,fdcoloring);CHKERRA(ierr);
  }

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Customize nonlinear solver; set runtime options
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  /*
     Set runtime options (e.g., -snes_monitor -snes_rtol <rtol> -ksp_type <type>)
  */
  ierr = SNESSetFromOptions(snes);CHKERRA(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Evaluate initial guess; then solve nonlinear system
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /*
     Note: The user should initialize the vector, x, with the initial guess
     for the nonlinear solver prior to calling SNESSolve().  In particular,
     to employ an initial guess of zero, the user should explicitly set
     this vector to zero by calling VecSet().
  */
  ierr = FormInitialGuess(&user,x);CHKERRA(ierr);
  ierr = SNESSolve(snes,x,&its);CHKERRA(ierr); 

  ierr = PetscPrintf(PETSC_COMM_WORLD,"Number of Newton iterations = %d\n",its);CHKERRA(ierr);

  /*
       Visualize the solution
  */
  ierr = VecView(x,VIEWER_DRAW_WORLD);CHKERRA(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Free work space.  All PETSc objects should be destroyed when they
     are no longer needed.
   - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = MatDestroy(J);CHKERRA(ierr);
  ierr = MatFDColoringDestroy(fdcoloring);CHKERRA(ierr);  
  ierr = VecDestroy(user.localX);CHKERRA(ierr); ierr = VecDestroy(x);CHKERRA(ierr);
  ierr = VecDestroy(user.localF);CHKERRA(ierr); ierr = VecDestroy(r);CHKERRA(ierr);      
  ierr = SNESDestroy(snes);CHKERRA(ierr);  ierr = DADestroy(user.da);CHKERRA(ierr);
  PetscFinalize();

  return 0;
}

/*
      Define macros to allow us to easily access the two components of the PDE
   solution.
*/
#define Psi(i)   2*(i)
#define Omega(i) 2*(i)+1

/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "FormInitialGuess"
/* 
   FormInitialGuess - Forms initial approximation.

   Input Parameters:
   user - user-defined application context
   X - vector

   Output Parameter:
   X - vector
 */
int FormInitialGuess(AppCtx *user,Vec X)
{
  int     i,j,row,ierr,xs,ys,xm,ym,gxm,gym,gxs,gys;
  Scalar  *x;
  Vec     localX = user->localX;

  /*
     Get a pointer to vector data.
       - For default PETSc vectors, VecGetArray() returns a pointer to
         the data array.  Otherwise, the routine is implementation dependent.
       - You MUST call VecRestoreArray() when you no longer need access to
         the array.
  */
  ierr = VecGetArray(localX,&x);CHKERRQ(ierr);

  /*
     Get local grid boundaries (for 2-dimensional DA):
       xs, ys   - starting grid indices (no ghost points)
       xm, ym   - widths of local grid (no ghost points)
       gxs, gys - starting grid indices (including ghost points)
       gxm, gym - widths of local grid (including ghost points)
  */
  ierr = DAGetCorners(user->da,&xs,&ys,PETSC_NULL,&xm,&ym,PETSC_NULL);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&gxs,&gys,PETSC_NULL,&gxm,&gym,PETSC_NULL);CHKERRQ(ierr);

  /*
     Compute initial guess over the locally owned part of the grid
  */
  for (j=ys; j<ys+ym; j++) {
    for (i=xs; i<xs+xm; i++) {
      row = i - gxs + (j - gys)*gxm; 
      x[Psi(row)]   = 1.0;
      x[Omega(row)] = 1.0;
    }
  }

  /*
     Restore vector
  */
  ierr = VecRestoreArray(localX,&x);CHKERRQ(ierr);

  /*
     Insert values into global vector
  */
  ierr = DALocalToGlobal(user->da,localX,INSERT_VALUES,X);CHKERRQ(ierr);
  return 0;
} 
/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "FormFunctionFast"
/* 
   FormFunctionFast - Evaluates the nonlinear function, F(x).
   This routine is a fast version of the user-provided function;
   see also FormFunctionSlow().

   Input Parameters:
.  snes - the SNES context
.  X - input vector
.  ptr - optional user-defined context, as set by SNESSetFunction()

   Output Parameter:
.  F - function vector

   Notes:
   We process the boundary nodes before handling the interior
   nodes, so that no conditional statements are needed within the
   double loop over the local grid indices.  This implementation
   is more efficient than FormFunctionSlow().
 */
int FormFunctionFast(SNES snes,Vec X,Vec F,void *ptr)
{
  AppCtx  *user = (AppCtx*)ptr;
  int     ierr,i,j,row,mx,my,xs,ys,xm,ym,gxs,gys,gxm,gym;
  int     xints,xinte,yints,yinte;
  double  two = 2.0,one = 1.0,hx,hy,hxdhy,hydhx;
  Scalar  u,uxx,uyy,*x,*f;
  Vec     localX = user->localX,localF = user->localF; 

  mx = user->mx;            my = user->my;            
  hx = one/(double)(mx-1);  hy = one/(double)(my-1);
  hxdhy = hx/hy;            hydhx = hy/hx;

  /*
     Scatter ghost points to local vector, using the 2-step process
        DAGlobalToLocalBegin(), DAGlobalToLocalEnd().
     By placing code between these two statements, computations can be
     done while messages are in transition.
  */
  ierr = DAGlobalToLocalBegin(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);
  ierr = DAGlobalToLocalEnd(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);

  /*
     Get pointers to vector data
  */
  ierr = VecGetArray(localX,&x);CHKERRQ(ierr);
  ierr = VecGetArray(localF,&f);CHKERRQ(ierr);

  /*
     Get local grid boundaries
  */
  ierr = DAGetCorners(user->da,&xs,&ys,PETSC_NULL,&xm,&ym,PETSC_NULL);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&gxs,&gys,PETSC_NULL,&gxm,&gym,PETSC_NULL);CHKERRQ(ierr);

  /*
     Compute function over the locally owned part of the grid
  */
  xints = xs; xinte = xs+xm; yints = ys; yinte = ys+ym;

  /* Test whether we are on the bottom edge of the global array */
  if (yints == 0) {
    yints = yints + 1;
    /* bottom edge */
    row = xs - gxs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
        f[Psi(row)]   = u = x[Psi(row)];
        f[Omega(row)]   = (u - x[Psi(row+gxm)])*hx; 
/*        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row+gxm)] + x[Psi(row+2*gxm)])*hxdhy; */
    }
  }

  /* Test whether we are on the top edge of the global array */
  if (yinte == my) {
    yinte = yinte - 1;
    /* top edge */
    row = (ys + ym - 1 - gys)*gxm + xs - gxs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
	f[Psi(row)]   = u = x[Psi(row)];
        f[Omega(row)]   = (u - x[Psi(row-gxm)])*hx - user->lidvelocity*hx*hy;
/*        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row-gxm)] + x[Psi(row-2*gxm)])*hxdhy; */
    }
  }

  /* Test whether we are on the left edge of the global array */
  if (xints == 0) {
    xints = xints + 1;
    /* left edge */
    for (j=ys; j<ys+ym; j++) {
      row = (j - gys)*gxm + xs - gxs; 
      f[Psi(row)]   = u  = x[Psi(row)];
      f[Omega(row)] = (u - x[Psi(row+1)])*hy;
/*      f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row+1)] + x[Psi(row+2)])*hydhx;  */
    }
  }

  /* Test whether we are on the right edge of the global array */
  if (xinte == mx) {
    xinte = xinte - 1;
    /* right edge */ 
    for (j=ys; j<ys+ym; j++) {
      row = (j - gys)*gxm + xs + xm - gxs - 1; 
      f[Psi(row)]   = u  = x[Psi(row)];
      f[Omega(row)]  = (u - x[Psi(row-1)])*hy;
/*      f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row-1)] + x[Psi(row-2)])*hydhx; */
    }
  }

  /* Compute over the interior points */
  for (j=yints; j<yinte; j++) {
    row = (j - gys)*gxm + xints - gxs - 1; 
    for (i=xints; i<xinte; i++) {
      row++;
        u             = x[Psi(row)];
        uxx           = (two*u - x[Psi(row-1)] - x[Psi(row+1)])*hydhx;
        uyy           = (two*u - x[Psi(row-gxm)] - x[Psi(row+gxm)])*hxdhy;
        f[Psi(row)]   = uxx + uyy + x[Omega(row)];
        u             = x[Omega(row)];
        uxx           = (two*u - x[Omega(row-1)] - x[Omega(row+1)])*hydhx;
        uyy           = (two*u - x[Omega(row-gxm)] - x[Omega(row+gxm)])*hxdhy;
        f[Omega(row)] = uxx + uyy + .25*(
                (x[Psi(row+gxm)] - x[Psi(row-gxm)])*(x[Omega(row+1)] - x[Omega(row-1)])
                - (x[Psi(row+1)] - x[Psi(row-1)])*(x[Omega(row+gxm)] - x[Omega(row-gxm)]));
    }
  }

  /*
     Restore vectors
  */
  ierr = VecRestoreArray(localX,&x);CHKERRQ(ierr);
  ierr = VecRestoreArray(localF,&f);CHKERRQ(ierr);

  /*
     Insert values into global vector
  */
  ierr = DALocalToGlobal(user->da,localF,INSERT_VALUES,F);CHKERRQ(ierr);
  ierr = PLogFlops(11*ym*xm);CHKERRQ(ierr);

  return 0; 
} 
/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "FormJacobian"
/*
   FormJacobian - Evaluates Jacobian matrix.

   Input Parameters:
.  snes - the SNES context
.  x - input vector
.  ptr - optional user-defined context, as set by SNESSetJacobian()

   Output Parameters:
.  A - Jacobian matrix
.  B - optionally different preconditioning matrix
.  flag - flag indicating matrix structure

   Notes:
   Due to grid point reordering with DAs, we must always work
   with the local grid points, and then transform them to the new
   global numbering with the local-to-global mapping.  We cannot work
   directly with the global numbers for the original uniprocessor grid!  

   Two methods are available for imposing this transformation
   when setting matrix entries:
     (A) MatSetValuesLocal(), using the local ordering (including
         ghost points!)
         - Do the following two steps once, before calling SNESSolve()
           - Use DAGetISLocalToGlobalMapping() to extract the
             local-to-global map from the DA
           - Associate this map with the matrix by calling
             MatSetLocalToGlobalMapping() 
         - Then set matrix entries using the local ordering
           by calling MatSetValuesLocal()
     (B) MatSetValues(), using the global ordering 
         - Use DAGetGlobalIndices() to extract the local-to-global map
         - Then apply this map explicitly yourself
         - Set matrix entries using the global ordering by calling
           MatSetValues()
   Option (A) seems cleaner/easier in many cases, and is the procedure
   used in this example.
*/
int FormJacobian(SNES snes,Vec X,Mat *J,Mat *B,MatStructure *flag,void *ptr)
{
  AppCtx  *user = (AppCtx*)ptr;  /* user-defined application context */
  Mat     jac = *B;                /* Jacobian matrix */
  Vec     localX = user->localX;   /* local vector */
  int     ierr,i,j,row = 0,mx,my,col[5],ROW;
  int     xs,ys,xm,ym,gxs,gys,gxm,gym;
  Scalar  one = 1.0,v[5],hx,hy,hxdhy,*x;
  int     xints,xinte,yints,yinte;

  mx = user->mx;            my = user->my;            
  hx = one/(double)(mx-1);  hy = one/(double)(my-1);
  hxdhy = hx/hy;

  /******************************************************************
  *******************************************************************
       Form analytic Jacobian. NOT YET DONE!!!  
       The analytic Jacobian must be coded; this is just the function
       evaluation routine.
  *******************************************************************
  *******************************************************************/

  /*
     Scatter ghost points to local vector, using the 2-step process
        DAGlobalToLocalBegin(), DAGlobalToLocalEnd().
     By placing code between these two statements, computations can be
     done while messages are in transition.
  */
  ierr = DAGlobalToLocalBegin(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);
  ierr = DAGlobalToLocalEnd(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);

  /*
     Get pointers to vector data
  */
  ierr = VecGetArray(localX,&x);CHKERRQ(ierr);

  /*
     Get local grid boundaries
  */
  ierr = DAGetCorners(user->da,&xs,&ys,PETSC_NULL,&xm,&ym,PETSC_NULL);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&gxs,&gys,PETSC_NULL,&gxm,&gym,PETSC_NULL);CHKERRQ(ierr);

  /*
     Compute function over the locally owned part of the grid
  */
  xints = xs; xinte = xs+xm; yints = ys; yinte = ys+ym;

  /* Test whether we are on the bottom edge of the global array */
  if (yints == 0) {
    yints = yints + 1;
    /* bottom edge */
    row = xs - gxs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
      v[0] = 1.0;
      ROW = Psi(row);
      ierr = MatSetValues(jac,1,&ROW,1,&ROW,v,INSERT_VALUES);CHKERRQ(ierr);
      ROW    = Omega(row);
      col[0] = Omega(row); col[1] = Psi(row); col[2] =  Psi(row+gxm); col[3] = Psi(row+2*gxm);
      v[0]   = hy*hx;      v[1]   = hxdhy;    v[2]   =  -2.0*hxdhy;   v[3]   = -hxdhy;
      ierr = MatSetValues(jac,1,&ROW,4,col,v,INSERT_VALUES);CHKERRQ(ierr);
    }
  }

  /* Test whether we are on the top edge of the global array */
  if (yinte == my) {
    yinte = yinte - 1;
    /* top edge */
    row = (ys + ym - 1 - gys)*gxm + xs - gxs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
    }
  }

  /* Test whether we are on the left edge of the global array */
  if (xints == 0) {
    xints = xints + 1;
    /* left edge */
    for (j=ys; j<ys+ym; j++) {
      row = (j - gys)*gxm + xs - gxs; 
      v[0] = 1.0;
      ROW = Psi(row);
      ierr = MatSetValues(jac,1,&ROW,1,&ROW,v,INSERT_VALUES);CHKERRQ(ierr);
    }
  }

  /* Test whether we are on the right edge of the global array */
  if (xinte == mx) {
    xinte = xinte - 1;
    /* right edge */ 
    for (j=ys; j<ys+ym; j++) {
      v[0] = 1.0;
      ROW = Psi(row);
      ierr = MatSetValues(jac,1,&ROW,1,&ROW,v,INSERT_VALUES);CHKERRQ(ierr);
    }
  }

  /* Compute over the interior points */
  for (j=yints; j<yinte; j++) {
    row = (j - gys)*gxm + xints - gxs - 1; 
    for (i=xints; i<xinte; i++) {
      row++;
                                        
    }
  }

  ierr = MatAssemblyBegin(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = VecRestoreArray(localX,&x);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(jac,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);

  *flag = SAME_NONZERO_PATTERN;
  return 0;
}

/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "FormFunctionSlow"
/* 
   FormFunctionSlow - Evaluates the nonlinear function, F(x).
   This routine is a slow version of the user-provided function; 
   see also FormFunctionFast().

   Input Parameters:
.  snes - the SNES context
.  X - input vector
.  ptr - optional user-defined context, as set by SNESSetFunction()

   Output Parameter:
.  F - function vector

   Notes:
   Here we process the boundary and interior nodes within a single
   double loop, so that conditional statements are required within
   the loops to test for boundary locations. A more efficient
   implementation is FormFunctionFast().
 */
int FormFunctionSlow(SNES snes,Vec X,Vec F,void *ptr)
{
  AppCtx  *user = (AppCtx*)ptr;
  int     ierr,i,j,row,mx,my,xs,ys,xm,ym,gxs,gys,gxm,gym;
  double  two = 2.0,one = 1.0,hx,hy,hxdhy,hydhx;
  Scalar  u,uxx,uyy,*x,*f;
  Vec     localX = user->localX,localF = user->localF; 

  mx = user->mx;            my = user->my;            
  hx = one/(double)(mx-1);  hy = one/(double)(my-1);
  hxdhy = hx/hy;            hydhx = hy/hx;

  /*
     Scatter ghost points to local vector, using the 2-step process
        DAGlobalToLocalBegin(), DAGlobalToLocalEnd().
     By placing code between these two statements, computations can be
     done while messages are in transition.
  */
  ierr = DAGlobalToLocalBegin(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);
  ierr = DAGlobalToLocalEnd(user->da,X,INSERT_VALUES,localX);CHKERRQ(ierr);

  /*
     Get pointers to vector data
  */
  ierr = VecGetArray(localX,&x);CHKERRQ(ierr);
  ierr = VecGetArray(localF,&f);CHKERRQ(ierr);

  /*
     Get local grid boundaries
  */
  ierr = DAGetCorners(user->da,&xs,&ys,PETSC_NULL,&xm,&ym,PETSC_NULL);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(user->da,&gxs,&gys,PETSC_NULL,&gxm,&gym,PETSC_NULL);CHKERRQ(ierr);

  /*
     Compute function over the locally owned part of the grid
  */
  for (j=ys; j<ys+ym; j++) {
    row = (j - gys)*gxm + xs - gxs - 1; 
    for (i=xs; i<xs+xm; i++) {
      row++;
      /* left edge */
      if (i == 0) {

        f[Psi(row)]   = u  = x[Psi(row)];
        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row+1)] + x[Psi(row+2)])*hydhx; 

      /* right edge */
      } else if (i == mx-1) {
    
        f[Psi(row)]   = u  = x[Psi(row)];
        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row-1)] + x[Psi(row-2)])*hydhx; 

      /* bottom edge */
      } else if (j == 0) {

        f[Psi(row)]   = u = x[Psi(row)];
        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row+gxm)] + x[Psi(row+2*gxm)])*hxdhy; 

      /* top edge */
      } else if (j == my - 1) {
      
        u             = x[Psi(row)];
        f[Psi(row)]   = (u - x[Psi(row-gxm)])*hx - user->lidvelocity*hx*hy;
        f[Omega(row)] = x[Omega(row)]*hy*hx - (u - 2.0*x[Psi(row-gxm)] + x[Psi(row-2*gxm)])*hxdhy;

      /* interior */
      } else {

        u             = x[Psi(row)];
        uxx           = (two*u - x[Psi(row-1)] - x[Psi(row+1)])*hydhx;
        uyy           = (two*u - x[Psi(row-gxm)] - x[Psi(row+gxm)])*hxdhy;
        f[Psi(row)]   = uxx + uyy + x[Omega(row)];
        u             = x[Omega(row)];
        uxx           = (two*u - x[Omega(row-1)] - x[Omega(row+1)])*hydhx;
        uyy           = (two*u - x[Omega(row-gxm)] - x[Omega(row+gxm)])*hxdhy;
        f[Omega(row)] = uxx + uyy + .25*(
                (x[Psi(row+gxm)] - x[Psi(row-gxm)])*(x[Omega(row+1)] - x[Omega(row-1)])
                - (x[Psi(row+1)] - x[Psi(row-1)])*(x[Omega(row+gxm)] - x[Omega(row-gxm)]));
                                        
      }
    }
  }

  /*
     Restore vectors
  */
  ierr = VecRestoreArray(localX,&x);CHKERRQ(ierr);
  ierr = VecRestoreArray(localF,&f);CHKERRQ(ierr);

  /*
     Insert values into global vector
  */
  ierr = DALocalToGlobal(user->da,localF,INSERT_VALUES,F);CHKERRQ(ierr);
  ierr = PLogFlops(11*ym*xm);CHKERRQ(ierr);
  return 0; 
} 


/* ------------------------------------------------------------------------------*/
#undef __FUNC__  
#define __FUNC__ "DAGetColoringSpecial" 
/*
    DAGetColoringSpecial - Gets the coloring required for computing
    the Jacobian via finite differences on a function defined using the
    standard nine-point stencil on a two-dimensional regular grid, plus
    a three-point stencil along the boundary. 

    This routine is an extension of DAGetColoring() to support a
    double-width stencil along the problem domain's boundary.  This boundary
    stencil requires the use of additional colors in the sparse finite
    difference approximation of the Jacobian matrix compared to the colors
    that would be needed for the problem's interior; also, this stencil
    correspondingly introduces more nonzero Jacobian entries.

    Note that this routine generates more fill than is actually required
    for this particular problem, since it assumes a full nine-point stencil
    coupling for both components.  One could be more careful below to
    generate just the absolute non-zero pattern.

    Input Parameter:
.   da - the distributed array

    Output Parameters:
.   coloring - matrix coloring for use in finite difference Jacobian
               approximations
.   J  - matrix with the correct nonzero structure
         (but obviously without the correct Jacobian values)

.seealso: DAGetColoring()

*/
int DAGetColoringSpecial(DA da,ISColoring *coloring,Mat *J)
{
  int      ierr,xs,ys,nx,ny,*colors,i,j,ii,slot,gxs,gys,ys1,ny1,nx1,gnx,gny;           
  int      m,n,dim,w,s,*indices,*gindices,k,xs1,nc,ng,*cols;
  MPI_Comm comm;
  Scalar   *values;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

       This figure illustrates the global problem domain and the
       locally owned region for a 2-dimensional regular grid.

                                  m
          ------------------------------------------------------
         |                                                     |
         |                                                     |
         |               ----------------------                |
         |               |                    |                |
      n  |           yn  |                    |                |
         |               |                    |                |
         |               .---------------------                |
         |             (xs,ys)     xn                          |
         |            .                                        |
         |         (gxs,gys)                                   |
         |                                                     |
          -----------------------------------------------------

          xs,ys   - local starting indices (no ghost points)
          xn,yn   - widths of local grid (no ghost points)
          gxs,gys - local starting indices (including ghost points)
          gxn,gyn - widths of local grid (including ghost points)

     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

  ierr = DAGetInfo(da,&dim,&m,&n,0,0,0,0,&w,&s,0,0);CHKERRQ(ierr);
  if (dim != 2) SETERRQ(1,0,"2d only");
  if (s != 1)   SETERRQ(1,0,"Stencil width 1 only");

  /* also no support for periodic boundary conditions */
  nc      = w;
  values  = (Scalar*)PetscMalloc(9*nc*nc*sizeof(Scalar));CHKPTRQ(values);
  ierr    = PetscMemzero(values,9*nc*nc*sizeof(Scalar));CHKERRQ(ierr);
  cols    = (int*)PetscMalloc(nc*sizeof(int));CHKPTRQ(cols);
  indices = (int*)PetscMalloc(9*nc*nc*sizeof(int));CHKPTRQ(cols);

  ierr = DAGetCorners(da,&xs,&ys,0,&nx,&ny,0);CHKERRQ(ierr);
  ierr = DAGetGhostCorners(da,&gxs,&gys,0,&gnx,&gny,0);CHKERRQ(ierr);
  ierr = PetscObjectGetComm((PetscObject)da,&comm);CHKERRQ(ierr);

  /* Create the coloring */
  colors = (int*)PetscMalloc(nc*nx*ny*sizeof(int));CHKPTRQ(colors);
  ii = 0;
  for (j=ys; j<ys+ny; j++) {
    for (i=xs; i<xs+nx; i++) {
      for (k=0; k<nc; k++) {
        colors[ii++] = k + nc*((i % 4) + 4*(j % 4));
      }
    }
  }
  ierr = ISColoringCreate(comm,nc*nx*ny,colors,coloring);CHKERRQ(ierr);
  ierr = PetscFree(colors);CHKERRQ(ierr);

  /* Create an empty Jacobian matrix */
  ierr = MatCreateMPIAIJ(comm,nc*nx*ny,nc*nx*ny,PETSC_DECIDE,PETSC_DECIDE,
                         9*nc,0,0,0,J);CHKERRQ(ierr);  

  ierr = DAGetGlobalIndices(da,&ng,&gindices);CHKERRQ(ierr);
  {
    ISLocalToGlobalMapping ltog;
    ierr = ISLocalToGlobalMappingCreate(PETSC_COMM_SELF,ng,gindices,&ltog);CHKERRQ(ierr);
    ierr = MatSetLocalToGlobalMapping(*J,ltog);CHKERRQ(ierr);
    ierr = ISLocalToGlobalMappingDestroy(ltog);CHKERRQ(ierr);
  }

  /*
    For each node in the grid: we get the neighbors in the local, 
    on-processor ordering that includes the ghost points; then
    MatSetValuesLocal() maps these indices to the global PETSc ordering.
  */
  
  /* Fill up Jacobian for left edge */
  if (xs == 0) {
    ys1 = ys;
    ny1 = ny;
    /* lower left corner */
    if (ys == 0) {
      ys1++;
      ny1--;
      slot = xs - gxs + gnx*(ys - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot);         
        indices[1*nc+k] = k + nc*(slot + 1); 
        indices[2*nc+k] = k + nc*(slot + 2); 
        indices[3*nc+k] = k + nc*(slot + gnx);  
        indices[4*nc+k] = k + nc*(slot + gnx + 1);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,5*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
    /* upper left corner */
    if (ys + ny == n) {
      ny1--;
      slot = xs - gxs + gnx*(ys + (ny - 1) - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - gnx);  
        indices[1*nc+k] = k + nc*(slot - gnx + 1);
        indices[2*nc+k] = k + nc*(slot);         
        indices[3*nc+k] = k + nc*(slot + 1);
        indices[4*nc+k] = k + nc*(slot + 2);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,5*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
    for (j=ys1; j<ys1+ny1; j++) {
      slot = xs - gxs + gnx*(j - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - gnx);  
        indices[1*nc+k] = k + nc*(slot - gnx + 1);
        indices[2*nc+k] = k + nc*(slot);         
        indices[3*nc+k] = k + nc*(slot + 1);
        indices[4*nc+k] = k + nc*(slot + 2);
        indices[5*nc+k] = k + nc*(slot + gnx);   
        indices[6*nc+k] = k + nc*(slot + gnx + 1);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,7*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);  
    }
  }      

  /* Fill up Jacobian for right edge */
  if (xs + nx == m) {
    ys1 = ys;
    ny1 = ny;
    /* lower right corner */
    if (ys == 0) {
      ys1++;
      ny1--;
      slot = xs + (nx - 1) - gxs + gnx*(ys - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - 2);        
        indices[1*nc+k] = k + nc*(slot - 1);        
        indices[2*nc+k] = k + nc*(slot); 
        indices[3*nc+k] = k + nc*(slot + gnx - 1);  
        indices[4*nc+k] = k + nc*(slot + gnx);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,5*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
    /* upper right corner */
    if (ys + ny == n) {
      ny1--;
      slot = xs + (nx - 1) - gxs + gnx*(ys + (ny - 1) - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - gnx - 1);  
        indices[1*nc+k] = k + nc*(slot - gnx);
        indices[2*nc+k] = k + nc*(slot-2);        
        indices[3*nc+k] = k + nc*(slot-1);         
        indices[4*nc+k] = k + nc*(slot);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,5*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
    for (j=ys1; j<ys1+ny1; j++) {
      slot = xs + (nx - 1) - gxs + gnx*(j - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - gnx - 1);  
        indices[1*nc+k] = k + nc*(slot - gnx);
        indices[2*nc+k] = k + nc*(slot - 2);       
        indices[3*nc+k] = k + nc*(slot - 1);       
        indices[4*nc+k] = k + nc*(slot);
        indices[5*nc+k] = k + nc*(slot + gnx - 1); 
        indices[6*nc+k] = k + nc*(slot + gnx);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,7*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);   
    }
  }   

  /* Fill up Jacobian for bottom */
  if (ys == 0) {
    if (xs == 0) {nx1 = nx - 1; xs1 = 1;} else {nx1 = nx; xs1 = xs;}
    if (xs + nx == m) {nx1--;}
    for (i=xs1; i<xs1+nx1; i++) {
      slot = i - gxs + gnx*(ys - gys);
      for (k=0; k<nc; k++) {
        indices[0+k]    = k + nc*(slot - 1);      
        indices[1*nc+k] = k + nc*(slot);       
        indices[2*nc+k] = k + nc*(slot + 1);
        indices[3*nc+k] = k + nc*(slot + gnx - 1); 
        indices[4*nc+k] = k + nc*(slot + gnx);
        indices[5*nc+k] = k + nc*(slot + 2*gnx);
        indices[6*nc+k] = k + nc*(slot + gnx + 1);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,7*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
  }

  /* Fill up Jacobian for top */
  if (ys + ny == n) {
    if (xs == 0) {nx1 = nx - 1; xs1 = 1;} else {nx1 = nx; xs1 = xs;}
    if (xs + nx == m) {nx1--;}
    for (i=xs1; i<xs1+nx1; i++) {
      slot = i - gxs + gnx*(ys + (ny - 1) - gys);
      for (k=0; k<nc; k++) {    
        indices[0+k]    = k + nc*(slot - gnx - 1);
        indices[1*nc+k] = k + nc*(slot - 2*gnx);  
        indices[2*nc+k] = k + nc*(slot - gnx);  
        indices[3*nc+k] = k + nc*(slot - gnx + 1);
        indices[4*nc+k] = k + nc*(slot - 1);      
        indices[5*nc+k] = k + nc*(slot);       
        indices[6*nc+k] = k + nc*(slot  + 1);
        cols[k]         = k + nc*(slot);
      }
      ierr = MatSetValuesLocal(*J,nc,cols,7*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
  }

  /* Fill up Jacobian for interior of grid */
  nx1 = nx;
  ny1 = ny;
  if (xs == 0)    {xs++; nx1--;}
  if (ys == 0)    {ys++; ny1--;}
  if (xs+nx1 == m) {nx1--;}
  if (ys+ny1 == n) {ny1--;}
  for (i=xs; i<xs+nx1; i++) {
    for (j=ys; j<ys+ny1; j++) {
      slot = i - gxs + gnx*(j - gys);
      for (k=0; k<nc; k++) {    
        indices[0+k]    = k + nc*(slot - gnx - 1);
        indices[1*nc+k] = k + nc*(slot - gnx);
        indices[2*nc+k] = k + nc*(slot - gnx + 1);
        indices[3*nc+k] = k + nc*(slot - 1);  
        indices[4*nc+k] = k + nc*(slot);      
        indices[5*nc+k] = k + nc*(slot + 1);
        indices[6*nc+k] = k + nc*(slot + gnx - 1);
        indices[7*nc+k] = k + nc*(slot + gnx); 
        indices[8*nc+k] = k + nc*(slot + gnx + 1);
        cols[k]         = k + nc*(slot);
    }
      ierr = MatSetValuesLocal(*J,nc,cols,9*nc,indices,values,INSERT_VALUES);CHKERRQ(ierr);
    }
  }
  ierr = PetscFree(values);CHKERRQ(ierr);
  ierr = PetscFree(cols);CHKERRQ(ierr);
  ierr = PetscFree(indices);CHKERRQ(ierr);
  ierr = MatAssemblyBegin(*J,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);  
  ierr = MatAssemblyEnd(*J,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);  
  return 0;
}










