/*$Id: ex9.c,v 1.19 2000/09/22 20:46:14 bsmith Exp $*/

static char help[] = "Solves a nonlinear system in parallel with SNES.\n\
  \n\
The 2D driven cavity problem is solved in a velocity-vorticity formulation.\n\
The flow can be driven with the lid or with bouyancy or both:\n\
  -lidvelocity <lid>, where <lid> = dimensionless velocity of lid\n\
  -grashof <gr>, where <gr> = dimensionless temperature gradient\n\
  -prandtl <pr>, where <pr> = dimensionless thermal/momentum diffusity ratio\n\
Mesh sequencing is available, starting coarse and recursively doubling:\n\
  -mx <xg>, where <xg> = initial number of grid points in the x-direction\n\
  -my <yg>, where <yg> = initial number of grid points in the y-direction\n\
  -nlevels <nlevels>, where <nlevels> = number of refinement levels\n\
  -printg : print grid information\n\
Graphics of the contours of (U,V,Omega,T) are available on each grid:\n\
  -contours : draw contour plots of solution\n\
Parallelism can be invoked based on the DA construct:\n\
  -Nx <npx>, where <npx> = number of processors in the x-direction\n\
  -Ny <npy>, where <npy> = number of processors in the y-direction\n\n";

/*T
   Concepts: SNES^parallel multicomponent example
   Concepts: DA^using distributed arrays;
   Concepts: structured grid
   Concepts: multicomponent
   Processors: n
T*/

/* ------------------------------------------------------------------------

    We thank David E. Keyes for contributing the driven cavity discretization
    within this example code.

    This example incorporates mesh sequencing; ex8.c solves the problem for
    a single grid; additional files needed:
        common8and9.c - initial guess and nonlinear function evaluation 
                        routines (used by both ex8.c and ex9.c)
        ex8and9.h     - include file used by ex8.c and ex9.c

    This problem is modeled by the partial differential equation system
  
	- Lap(U) - Grad_y(Omega) = 0
	- Lap(V) + Grad_x(Omega) = 0
	- Lap(Omega) + Div([U*Omega,V*Omega]) - GR*Grad_x(T) = 0
	- Lap(T) + PR*Div([U*T,V*T]) = 0

    in the unit square, which is uniformly discretized in each of x and
    y in this simple encoding.

    No-slip, rigid-wall Dirichlet conditions are used for [U,V].
    Dirichlet conditions are used for Omega, based on the definition of
    vorticity: Omega = - Grad_y(U) + Grad_x(V), where along each
    constant coordinate boundary, the tangential derivative is zero.
    Dirichlet conditions are used for T on the left and right walls,
    and insulation homogeneous Neumann conditions are used for T on
    the top and bottom walls. 

    A finite difference approximation with the usual 5-point stencil 
    is used to discretize the boundary value problem to obtain a 
    nonlinear system of equations.  Upwinding is used for the divergence
    (convective) terms and central for the gradient (source) terms.
    
    The Jacobian can be either
      * formed via finite differencing using coloring (the default), or
      * applied matrix-free via the option -snes_mf 
        (for larger grid problems this variant may not converge 
        without a preconditioner due to ill-conditioning).

    Mesh sequencing can be used to move from a given initial discretization
    to successively finer ones, doubling the number of intervals in
    the x- and y-directions at each level.  The converged solution on
    one mesh is interpolated to form the starting guess for the next grid.
    The option -nlevels indicates the number of levels (default is 1).


  ------------------------------------------------------------------------- */

/* 
   Include "petscda.h" so that we can use distributed arrays (DAs).
   Include "petscsnes.h" so that we can use SNES solvers.  Note that this
   file automatically includes:
     petsc.h       - base PETSc routines   petscvec.h - vectors
     petscsys.h    - system routines       petscmat.h - matrices
     petscis.h     - index sets            petscksp.h - Krylov subspace methods
     petscviewer.h - viewers               petscpc.h  - preconditioners
     petscsles.h   - linear solvers       
*/
#include "petscsnes.h"
#include "petscda.h"

/* 
   User-defined application context - contains data needed by the 
   application-provided call-back routine, FormFunction(). 
*/
typedef struct {
   double     lidvelocity,prandtl,grashof; 
   int        mx,my;            /* discretization in x, y directions */
   int        mc;                /* components in unknown vector */
   Vec        localX,localF;    /* ghosted local vector */
   DA         da;                /* distributed array data structure (unknowns) */
   int        rank;              /* processor rank */
   int        size;              /* number of processors */
   MPI_Comm   comm;              /* MPI communicator */
   int        ilevel,nlevels;   /* current/total levels through problem */
   Vec        x_old;             /* old solution vector */
   char       **label;           /* labels for components */
   PetscTruth print_grid;        /* flag - 1 indicates printing grid info */
   PetscTruth print_vecs;        /* flag - 1 indicates printing vectors */
   PetscTruth draw_contours;     /* flag - 1 indicates drawing contours */
} AppCtx;

/* 
   User-defined routines
*/
extern int FormInitialGuess(AppCtx*,Vec);
extern int FormFunction(SNES,Vec,Vec,void*);
extern int UserLevelEnd(AppCtx*,Vec);
extern int InitializeProblem(int,AppCtx*,Vec*);

#undef __FUNC__
#define __FUNC__ "main"
int main(int argc,char **argv)
{
  SNES          snes;                /* nonlinear solver */
  Vec           x,r;                /* solution, residual vectors */
  Mat           J;                   /* Jacobian matrix */
  AppCtx        user;                /* user-defined work context */
  int           its;                 /* iterations for convergence */
  MatFDColoring fdcoloring;          /* matrix coloring context */
  int           dim,k,ierr;

  PetscInitialize(&argc,&argv,(char *)0,help);
  ierr = MPI_Comm_rank(PETSC_COMM_WORLD,&user.rank);CHKERRA(ierr);
  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&user.size);CHKERRA(ierr);
  user.comm = PETSC_COMM_WORLD;

  /* Repeatedly solve nonlinear problem, refining discretization each time */
  user.nlevels = 1;
  ierr = OptionsGetInt(PETSC_NULL,"-nlevels",&user.nlevels,PETSC_NULL);CHKERRA(ierr);
  for (k=0; k<user.nlevels; k++) {

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
      Create user context, set problem data, create vector data structures.
      Also, compute the initial guess.
      - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    ierr = InitializeProblem(k,&user,&x);CHKERRA(ierr);
    ierr = VecDuplicate(x,&r);CHKERRA(ierr);
    ierr = VecDuplicate(user.localX,&user.localF);CHKERRA(ierr);

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Create nonlinear solver context
       - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    ierr = SNESCreate(PETSC_COMM_WORLD,SNES_NONLINEAR_EQUATIONS,&snes);CHKERRA(ierr);
    ierr = VecGetSize(x,&dim);CHKERRA(ierr);
    ierr = PetscPrintf(user.comm,"level = %d, global size = %d, lid velocity = %g, prandtl # = %g, grashof # = %g\n",
       k+1,dim,user.lidvelocity,user.prandtl,user.grashof);CHKERRA(ierr);


    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Set function evaluation routine and vector
       - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    ierr = SNESSetFunction(snes,r,FormFunction,&user);CHKERRA(ierr);

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Create matrix data structure; set Jacobian evaluation routine
       - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    /* 
         Set up coloring information needed for sparse finite difference
         approximation of the Jacobian
     */
    {
    ISColoring iscoloring;
    ierr = DAGetColoring(user.da,&iscoloring,&J);CHKERRQ(ierr);
    ierr = MatFDColoringCreate(J,iscoloring,&fdcoloring);CHKERRQ(ierr); 
    ierr = MatFDColoringSetFunction(fdcoloring,(int (*)(void))FormFunction,&user);CHKERRQ(ierr);
    ierr = MatFDColoringSetFromOptions(fdcoloring);CHKERRQ(ierr); 
    ierr = ISColoringDestroy(iscoloring);CHKERRQ(ierr);
    }

    /* 
       Set Jacobian matrix data structure and default Jacobian evaluation
       routine. User can override with:
  
       -snes_mf : matrix-free Newton-Krylov method with no preconditioning
       -snes_mf_operator : form preconditioning matrix as set by the user,
                           but use matrix-free approx for Jacobian-vector
                           products within Newton-Krylov method
  
    */
    ierr = SNESSetJacobian(snes,J,J,SNESDefaultComputeJacobianColor,fdcoloring);CHKERRA(ierr);

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Customize nonlinear solver; set runtime options
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    /*
       Set runtime options (e.g., -snes_monitor -snes_rtol <rtol> -ksp_type <type>)
    */
    ierr = SNESSetFromOptions(snes);CHKERRA(ierr);

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Solve the nonlinear system
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
    /*
       Note: The user should initialize the vector, x, with the initial guess
       for the nonlinear solver prior to calling SNESSolve().  In particular,
       to employ an initial guess of zero, the user should explicitly set
       this vector to zero by calling VecSet(). [Here we set the initial 
       guess in the routine InitializeProblem().]
    */
    ierr = SNESSolve(snes,x,&its);CHKERRA(ierr); 

    ierr = PetscPrintf(PETSC_COMM_WORLD,"Number of Newton iterations = %d\n",its);CHKERRA(ierr);

    /*
       Visualize solution
    */

    ierr = VecView(x,VIEWER_DRAW_WORLD);CHKERRA(ierr);

    /* 
       User-defined actions at conclusion of each nonlinear solve
     */
    ierr = UserLevelEnd(&user,x);CHKERRA(ierr);

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
       Free work space.  All PETSc objects should be destroyed when they
       are no longer needed.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

    ierr = MatDestroy(J);CHKERRA(ierr);
    ierr = MatFDColoringDestroy(fdcoloring);CHKERRA(ierr);  
    ierr = VecDestroy(user.localX);CHKERRA(ierr); ierr = VecDestroy(x);CHKERRA(ierr);
    ierr = VecDestroy(user.localF);CHKERRA(ierr); ierr = VecDestroy(r);CHKERRA(ierr);      
    ierr = SNESDestroy(snes);CHKERRA(ierr);  
  }

  ierr = PetscFree(user.label);CHKERRA(ierr);
  ierr = DADestroy(user.da);CHKERRA(ierr);

  PetscFinalize();
  return 0;
}

/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "InitializeProblem"
/* 
   InitializeProblem - Initializes the problem for a given grid size.
   This routine forms the DA and vector data structures, and also
   computes the starting solution guess for the nonlinear solver.

   Input Parameters:
   ilevel - level number
   user - user-defined application context

   Output Parameter:
   xvec - solution vector

   Notes:
   Mesh sequencing can be used to move from a given initial discretization
   to successively finer ones, where we double the number of intervals in
   the x- and y-directions at each level.  
     - first level: Create the initial DA, extract vectors, and compute an 
                           initial guess via the routine FormInitialGuess().
     - successive levels:  Refine the DA and vectors, and interpolate the
                           converged solution to form the starting guess
                           then next grid.
 */
int InitializeProblem(int ilevel,AppCtx *user,Vec *xvec)
{
  int        Nx,Ny;              /* number of processors in x- and y- directions */
  int        xs,xm,ys,ym,Nlocal,ierr;
  DA         da_ref;
  Vec        xv;
  Mat        interp;

  user->ilevel = ilevel;

  /* First time through only ... */
  if (ilevel == 0) {

    /*
       Initialize problem parameters
    */
    user->mx = 4; user->my = 4; 
    ierr = OptionsGetInt(PETSC_NULL,"-mx",&user->mx,PETSC_NULL);CHKERRQ(ierr);
    ierr = OptionsGetInt(PETSC_NULL,"-my",&user->my,PETSC_NULL);CHKERRQ(ierr);
    /*
       No. of components in the unknown vector and auxiliary vector
    */
    user->mc = 4;
    /* 
       Problem parameters (velocity of lid, prandtl, and grashof numbers)
    */
    user->lidvelocity = 1.0/(user->mx*user->my);
    ierr = OptionsGetDouble(PETSC_NULL,"-lidvelocity",&user->lidvelocity,PETSC_NULL);CHKERRQ(ierr);
    user->prandtl = 1.0;
    ierr = OptionsGetDouble(PETSC_NULL,"-prandtl",&user->prandtl,PETSC_NULL);CHKERRQ(ierr);
    user->grashof = 1.0;
    ierr = OptionsGetDouble(PETSC_NULL,"-grashof",&user->grashof,PETSC_NULL);CHKERRQ(ierr);
    ierr = OptionsHasName(PETSC_NULL,"-printv",&user->print_vecs);CHKERRQ(ierr);
    ierr = OptionsHasName(PETSC_NULL,"-printg",&user->print_grid);CHKERRQ(ierr);
    ierr = OptionsHasName(PETSC_NULL,"-contours",&user->draw_contours);CHKERRQ(ierr);

    /*
       Create distributed array (DA) to manage parallel grid and vectors
       for principal unknowns (x) and governing residuals (f)
    */
    Nx = PETSC_DECIDE; Ny = PETSC_DECIDE;
    ierr = OptionsGetInt(PETSC_NULL,"-Nx",&Nx,PETSC_NULL);CHKERRQ(ierr);
    ierr = OptionsGetInt(PETSC_NULL,"-Ny",&Ny,PETSC_NULL);CHKERRQ(ierr);
    ierr = DACreate2d(PETSC_COMM_WORLD,DA_NONPERIODIC,DA_STENCIL_STAR,user->mx,
                    user->my,Nx,Ny,user->mc,1,PETSC_NULL,PETSC_NULL,&user->da);CHKERRQ(ierr);
    /*
       Extract global vector from DA
    */
    ierr = DACreateGlobalVector(user->da,&xv);CHKERRQ(ierr);

    /* 
       Label PDE components
     */
    user->label = (char**)PetscMalloc(user->mc*sizeof(char*));CHKPTRQ(user->label);
    user->label[0] = "Velocity (U)";
    user->label[1] = "Velocity (V)";
    user->label[2] = "Omega";
    user->label[3] = "Temperature";

    user->x_old = 0;
  }

  /* Subsequent times through ... */
  else {

    /* Refine distributed array. */
    ierr = DARefine(user->da,&da_ref);CHKERRQ(ierr);
    ierr = DACreateGlobalVector(da_ref,&xv);CHKERRQ(ierr);

    /* interpolate vector to finer grid */
    ierr = DAGetInterpolation(user->da,da_ref,&interp,PETSC_NULL);CHKERRQ(ierr);
    ierr = MatInterpolate(interp,user->x_old,xv);CHKERRQ(ierr);
    ierr = MatDestroy(interp);CHKERRQ(ierr);

    /* free old da */
    ierr = DADestroy(user->da);CHKERRQ(ierr);
    user->da = da_ref;
    ierr = DAGetInfo(user->da,PETSC_NULL,&user->mx,&user->my,PETSC_NULL,PETSC_NULL,
           PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);
  }

  /* Get local vector */
  ierr = DACreateLocalVector(user->da,&user->localX);CHKERRQ(ierr);

  /* Print grid info */
  if (user->print_grid) {
    ierr = DAView(user->da,VIEWER_STDOUT_SELF);CHKERRQ(ierr);
    ierr = DAGetCorners(user->da,&xs,&ys,0,&xm,&ym,0);CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"global grid: %d X %d with %d components per node ==> global vector dimension %d\n",
      user->mx,user->my,user->mc,user->mc*user->mx*user->my);CHKERRQ(ierr); fflush(stdout);
    ierr = VecGetLocalSize(xv,&Nlocal);CHKERRQ(ierr);
    ierr = PetscSynchronizedPrintf(PETSC_COMM_WORLD,"[%d] local grid %d X %d with %d components per node ==> local vector dimension %d\n",user->rank,xm,ym,user->mc,Nlocal);CHKERRQ(ierr);
    ierr = PetscSynchronizedFlush(PETSC_COMM_WORLD);CHKERRQ(ierr);
  }  

  /* Compute initial guess first time only; otherwise, use refined vector */
  if (user->ilevel == 0)  {
    ierr = FormInitialGuess(user,xv);CHKERRQ(ierr);
  } else {

    /* Print refined vector */
    ierr = VecView(xv,VIEWER_DRAW_WORLD);CHKERRQ(ierr);

    ierr = DAGlobalToLocalBegin(user->da,xv,INSERT_VALUES,user->localX);CHKERRQ(ierr);
    ierr = DAGlobalToLocalEnd(user->da,xv,INSERT_VALUES,user->localX);CHKERRQ(ierr);

  }
  
  *xvec = xv;
  return 0;
}

/* ------------------------------------------------------------------- */
#undef __FUNC__
#define __FUNC__ "UserLevelEnd"
/* 
   UserLevelEnd - User-defined actions at conclusion of each nonlinear solve.

   Input Parameters:
   user - user-defined application context
   x - converged solution vector

   Notes:
   Here we save the solution vector for the next pass through the
   level.  Alternatively, we could write the vector in binary to a 
   file, stop the nonlinear solver, and then restart using this as 
   a new initial guess to be refined for solving a larger problem.
 */
int UserLevelEnd(AppCtx *user,Vec x)
{
  int ierr;

  /* Destroy the existing x_old vector if necessary */
  if (user->ilevel) {ierr = VecDestroy(user->x_old);CHKERRQ(ierr);}

  if (user->ilevel+1 != user->nlevels) {
    ierr = VecDuplicate(x,&user->x_old);CHKERRQ(ierr);
    ierr = VecCopy(x,user->x_old);CHKERRQ(ierr);
  }
  return 0;
}
